/* $Id: ArkLoaderMDL.cpp,v 1.40 2003/03/20 17:23:25 zongo Exp $
**
** Ark - Libraries, Tools & Programs for MMORPG developpements.
** Copyright (C) David Farrell <fdavid@cyberramp.net>
** Please see the file "AUTHORS" for a list of contributors
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sstream>

#include <stdio.h>

#include <Ark/Ark.h>
#include <Ark/ArkModel.h>
#include <Ark/ArkImage.h>
#include <Ark/ArkCache.h>
#include <Ark/ArkTexture.h>
#include <Ark/ArkModelBuild.h>
#include <Ark/ArkLoader.h>
#include <Ark/studio.h>


namespace Ark
{

   static studiohdr_t *LoadMDLFile (const String& name, Stream &fp)
   {
       // Gets size of file
       fp.seekg(0, std::ios::end);
       //size_t size = fp.Size();
       long size = fp.tellg();
       fp.seekg(0, std::ios::beg);

      // Try to get the size of the file
      if (size <= 0)
	 return NULL;

      uchar *pin = new uchar [size];
      if (!pin)
	 return NULL;
      fp.read ((char *) pin, size);

      // Check for a signature
      if (memcmp (pin, "IDST", 4) != 0 &&
	  memcmp (pin, "IDSQ", 4) != 0)
      {
	 char *c = (char *)pin;
	 Sys()->Warning ("  magic_number..error : no magic number "
			 "(%c%c%c%c)\n",
			 name.c_str(), c[0], c[1], c[2], c[3]);
	 delete[] pin;
	 return NULL;
      }
      
      return (studiohdr_t *)pin;
   }

   Vector3 VSwap (const Vector3 &v, const Matrix44 &tr)
   {
      //return tr.Transform (v);
      return v;
   }

   Vector3 VSwap (scalar X, scalar Y, scalar Z, const Matrix44 &tr)
   {
      //return tr.Transform (Vector3 (X, Y, Z));
      return Vector3 (X, Y, Z);
   }

   static String RemoveExtension(const String& str)
   {
       return str.substr(0, str.rfind('.'));
   }


   class ModelMDL
   {
      public:
	 studiohdr_t *m_pstudiohdr;

	 ModelMDL (studiohdr_t *hdr)
	 {
	    m_pstudiohdr = hdr;
	 }

	 Vector3 GetBonePos (mstudiobone_t *pbone,
			     mstudioanim_t *panim,
			     int frame)
	 {
	    mstudioanimvalue_t *av;
	    int j, k;
	    scalar pos[3];
	    
	    for (j = 0; j < 3; j++)
	    {
	       pos[j] = 0.0;
	       
	       if (panim->offset[j] != 0)
	       {
		  av = (mstudioanimvalue_t *)
		     ((uchar *)panim + panim->offset[j]);
		  k = frame;
		  
		  // find span of values that includes the frame we want
		  while (av->num.total <= k)
		  {
		     k -= av->num.total;
		     av += av->num.valid + 1;
		  }
		  
		  // if we're inside the span
		  if (av->num.valid > k)
		  {
		     pos[j] += av[k+1].value * pbone->scale[j];
		  }
		  else
		  {
		     pos[j] += av[av->num.valid].value * pbone->scale[j];
		  }
	       }
	    }

	    return Vector3 (pos[0], pos[1], pos[2]);
	 }

	 Vector3
	 GetBoneRot (mstudiobone_t *pbone,
		     mstudioanim_t *panim,
		     int frame)
	 {
	    scalar a[3];
	    mstudioanimvalue_t *av;
	    int j, k;
	    
	    for (j = 0; j < 3; j++)
	    {
	       if (panim->offset[j+3] == 0)
		  a[j] = pbone->value[j+3]; // default;
	       else
	       {
		  av = (mstudioanimvalue_t *)
		     ((uchar *)panim + panim->offset[j+3]);
		  
		  k = frame;
		  while (av->num.total <= k)
		  {
		     k -= av->num.total;
		     av += av->num.valid + 1;
		  }
		  
		  // Bah, missing blend!
		  if (av->num.valid > k)
		     a[j] = av[k+1].value;
		  else
		     a[j] = av[av->num.valid].value;
		  
		  a[j] = pbone->value[j+3] + a[j] * pbone->scale[j+3];
	       }

	       a[j] = Ark::Math::RadToDeg (a[j]);
	    }
	    
	    return Vector3 (a[2], a[1], a[0]);
	 }

	 void FillFrame (mstudioseqdesc_t *pseqdesc,
			 mstudioanim_t *panim,
			 int frame,
			 Keyframe &kf)
	 {
	    mstudiobone_t *pbone;
        
	    pbone = (mstudiobone_t *)
	       ((uchar *)m_pstudiohdr + m_pstudiohdr->boneindex);

	    // Compute the position and rotation of each bone
	    for (int i = 0; i < m_pstudiohdr->numbones; i++, pbone++, panim++)
	    {
	       BoneFrame bf;
	       bf.m_Position = GetBonePos(pbone, panim, frame);
	       bf.m_Rotation = GetBoneRot(pbone, panim, frame);

	       kf.push_back(bf);
	    }

	    if (pseqdesc->motiontype & STUDIO_X)
	       kf[pseqdesc->motionbone].m_Position.X = 0.0;
	    if (pseqdesc->motiontype & STUDIO_Y)
	       kf[pseqdesc->motionbone].m_Position.Y = 0.0;
	    if (pseqdesc->motiontype & STUDIO_Z)
	       kf[pseqdesc->motionbone].m_Position.Z = 0.0;

	    
	 }

	 mstudioseqdesc_t &GetSequence (int i)
	 {
	    return ((mstudioseqdesc_t *)
	       ((uchar *)m_pstudiohdr + m_pstudiohdr->seqindex))[i];
	 }

	 mstudioanim_t *
	 GetAnim (mstudioseqdesc_t *pseqdesc)
	 {
	    mstudioseqgroup_t *pseqgroup;
	    pseqgroup = (mstudioseqgroup_t *)
	       ((uchar *)m_pstudiohdr + m_pstudiohdr->seqgroupindex)
	       + pseqdesc->seqgroup;
	    
	    if (pseqdesc->seqgroup == 0)
	    {
	       return (mstudioanim_t *)
		  ((uchar *)m_pstudiohdr +
		   pseqgroup->data +
		   pseqdesc->animindex);
	    }
	    
	    return NULL;
	 }
	 
	 void FillSequence (Sequence *pseq, const String &name)
	 {
	    int n = 0;

	    for (int i = 0; i < m_pstudiohdr->numseq; i++)
	    {
	       mstudioseqdesc_t &desc = GetSequence(i);
	       if (name == desc.label)
	       { n = i; break; }
	    }

	    // cout << "Loading sequence " << name << " (" << n << ")\n";

	    mstudioseqdesc_t &desc = GetSequence(n);
	    mstudioanim_t *anim = GetAnim (&desc);

	    if (anim == NULL)
	       return;

	    Sequence &seq = *pseq;

	    int numframes = desc.numframes;
	    scalar fps = desc.fps;

#if 0
	    Sys()->Log ("seq : %d, %s, %d, %f\n", n, desc.label,
			numframes, fps);
#endif
	    seq.m_EndTime = scalar(numframes-1) / fps;
	    seq.m_FPS = fps;
	    seq.m_Keyframes.resize (numframes);

	    seq.m_BBox.AddPoint (desc.bbmin);
	    seq.m_BBox.AddPoint (desc.bbmax);

	    Matrix44 m; m.MakeRotationX(-90.f);
	    seq.m_BBox.Transform (m);

	    /*
	    Sys()->Log ("Sequence %s: %f fps, %d frames\n"
			"\tstart_frame : %d\n"
			"\tstart_time : %f\n",
			desc.label, fps, numframes,
			m_CurFrame, m_CurTime);
	    */

	    /// Copy frames
	    for (int f = 0; f < numframes; f++)
	    {
	       FillFrame (&desc, anim, f, seq.m_Keyframes[f]);
	    }	    
	 }

	 void FillSkeleton (Skeleton *skel, const String &fname)
	 {
	    int i;
	    skel->m_Bones.resize (m_pstudiohdr->numbones);
	    skel->ComputeBoneOrder();
	    
	    mstudiobone_t *bones = (mstudiobone_t*)
	       ((uchar*) (m_pstudiohdr) + m_pstudiohdr->boneindex);

	    for (i = 0; i < m_pstudiohdr->numbones; i++)
	    {
	       //Sys()->Log ("Adding bone %s..\n", bones [i].name);
	       Bone &bone = skel->m_Bones[i];
	       bone.m_Name = bones[i].name;
	       bone.m_Parent = bones[i].parent;
	       
	       // Defaults
	       bone.m_DefPosition =
		  Vector3 (bones[i].value[0],
			   bones[i].value[1],
			   bones[i].value[2]);
	       
	       bone.m_DefRotation =
		  Vector3 (Ark::Math::RadToDeg (bones[i].value[5]),
			   Ark::Math::RadToDeg (bones[i].value[4]),
			   Ark::Math::RadToDeg (bones[i].value[3]));

	       if (bone.m_Parent == -1)
	       {
		  /*
		  Quat bquat (bone.m_DefRotation.X,
			      bone.m_DefRotation.Y,
			      bone.m_DefRotation.Z);
		  Quat rot (-90.0, 0.0, 0.0);

		  (bquat * rot).GetEuler (&bone.m_DefRotation.X,
					  &bone.m_DefRotation.Y,
					  &bone.m_DefRotation.Z);
		  */
	       }
	              
	       // Scale factors
	       /*
	       bone.m_PositionScale =
		  Vector3 (bones[i].scale[0],
			   bones[i].scale[1],
			   bones[i].scale[2]);
	       bone.m_RotationScale =
		  Vector3 (bones[i].scale[5],
			   bones[i].scale[4],
			   bones[i].scale[3]);
	       */
	       
	       for (int j = 0; j < 6; j++)
		  bone.m_BoneControllers[j] = bones[i].bonecontroller[j];
	    }

	    // Copy hitboxes.. ==============================
	    skel->m_HitBoxes.resize (m_pstudiohdr->numhitboxes);
	    
	    mstudiobbox_t *boxes = (mstudiobbox_t*)
	       ((uchar*) (m_pstudiohdr) + m_pstudiohdr->hitboxindex);

	    for (i = 0; i < m_pstudiohdr->numhitboxes; i++)
	    {
	       HitBox &box = skel->m_HitBoxes[i];
	       box.m_Min = boxes[i].bbmin;
	       box.m_Max = boxes[i].bbmax;
	       box.m_Bone = boxes[i].bone;
	    }
	    
	    // Copy controllers.. ==============================
	    skel->m_Controllers.resize
	       (m_pstudiohdr->numbonecontrollers);

	    mstudiobonecontroller_t *conts =
	       (mstudiobonecontroller_t*)
	       ((uchar*) (m_pstudiohdr) + m_pstudiohdr->bonecontrollerindex);

	    for (i = 0; i < m_pstudiohdr->numbonecontrollers; i++)
	    {
	       BoneController &cont = skel->m_Controllers[i];
	       cont.m_Bone = conts[i].bone;
	       cont.m_Channel = conts[i].type;
	       cont.m_Start = conts[i].start;
	       cont.m_End = conts[i].end;
	       cont.m_Index = conts[i].index;
#if 0
	       Sys()->Log ("Controller %d: bone %d, channel %d, "
			   "start %f, end %f\n", 
			   i, 
			   cont.m_Bone, cont.m_Channel, cont.m_Start,
			   cont.m_End);
#endif
	    }

	    for (i = 0; i < m_pstudiohdr->numseq; i++)
	    {
	       mstudioseqdesc_t &desc = GetSequence(i);
	       
	       skel->m_Sequences[desc.label] = fname + "?" + desc.label;
	    }
	 }
   };

   class LoaderMDL : public Loader
   {
	 /// Returns true if the file pointed to by \c name does really
	 /// contains an HalfLife MDL, and if this loader can load it.
	 virtual bool CanLoad (ObjectType type, Stream &stream,
			       const String &name, const String &args)
	 {
	    if (type != V_MODEL && type != V_IMAGE && type != V_SEQUENCE)
	       return false;

	    uchar id[4];
	    stream.read ((char *)id, 4);

	    if (memcmp (id, "IDST", 4) != 0 &&
		memcmp (id, "IDSQ", 4) != 0)
	    {
	       return false;
	    }

	    return true;
	 }

	 /// Returns informations about the formats supported by this loader.
	 virtual String GetInformations()
	 {
	    return "Half-life Model";
	 }
	 
	 bool
	 UploadTexture(Image *img,
		       mstudiotexture_t *ptexture,
		       uchar *data, uchar *pal)
	 {
	    int i, j;;
	    uchar *tex, *out;

	    if (!img->SetFormat (ptexture->width,
				 ptexture->height,
				 Image::RGB_888))
	       return false;

	    out = img->m_Data;
	    tex = out = img->m_Data;
 
	    // Convert to 32bit RGB
	    for (i=0 ; i<ptexture->height ; i++)
	    {
	       for (j=0 ; j<ptexture->width; j++)
	       {
		  uchar *pix = &pal[*data++ * 3];
		  *out++ = *pix++;
		  *out++ = *pix++;
		  *out++ = *pix++;
	       }
	    }

	    return true;
	 }
	    
	 bool LoadImage (Image *image, Stream &file, 
		 const String &name, const String &args,
			 Progress *progress, int granularity)
	 {
	    if (args == "")
	       return false;

	    studiohdr_t *hdr = LoadMDLFile (name, file);
	    uchar *pin = (uchar*) hdr;

	    mstudiotexture_t *ptexture;
	    ptexture = (mstudiotexture_t *)(pin + hdr->textureindex);   

	    int idx;

	    if ((! StringToInt(args, idx) ) || (idx > hdr->numtextures))
	    {
	       Sys()->Warning ("%s?%s : Index %d is out of bounds",
			       name.c_str(), idx);
	       return false;	      
	    }

	    bool ret = UploadTexture (image,
				      &ptexture[idx],
				      pin + ptexture[idx].index,
				      pin + ptexture[idx].height *
				      ptexture[idx].width +
				      ptexture[idx].index);
	    
	    delete hdr;
	    return ret;
	 }

	 bool LoadSequence (Sequence *seq, Stream &file, 
			    const String &name, const String &args,
			    Progress *progress, int granularity)
	 {
	    studiohdr_t *hdr = LoadMDLFile (name, file);
	    ModelMDL mdlbuild (hdr);
	    mdlbuild.FillSequence(seq, args);
	    
	    delete hdr;
	    return true;
	 }

	 
	 /// Load the file pointed to by \c name, and read a object in it.
	 /// It will update the progress every \c granularity percents.
	 virtual bool Load (Object *vis, Stream &file, const String &name, const String &args,
			    Cache *cache,
			    Progress *progress, int granularity)
	 {
	    if (vis == NULL)
	       return false;

	    // Models and Images can be read from HalfLife MDL files
	    switch (vis->Type())
	    {
	       case V_MODEL:
		  break;
	       case V_IMAGE:
		  return LoadImage ((Image*) vis, file, name, args, progress,
				    granularity);
	       case V_SEQUENCE:
		  return LoadSequence ((Sequence*) vis, file, name, args,
				       progress, granularity);
	       default:
		  return false;
	    }
	    
	    studiohdr_t *hdr = LoadMDLFile (name, file);
	    uchar *pin = (uchar*) hdr;
	    
	    if (hdr == NULL)
	       return false;
	    
	    Model *mdl = (Model*) vis;
	    ModelMDL mdlbuild (hdr);
	    int i;

	    // Material directory : remove the .mdl extension
	    String matdir = RemoveExtension( name );
	    
	    // Copy bones.. ==============================
	    Skeleton *skel = new Skeleton (name);
	    mdl->m_Skeleton = skel;
	    mdlbuild.FillSkeleton (skel, name);	   

	    mstudiotexture_t *ptexture;
	    ptexture = (mstudiotexture_t *)(pin + hdr->textureindex);   

	    /// Copy skins..
	    for (i = 0; i < hdr->numtextures; i++)
	    {
	       // Try to get an external material, in the material directory..
	       String abs_mat_name = RemoveExtension( matdir + "/" + ptexture[i].name);
	       abs_mat_name += ".mat";
	       
	       MaterialPtr mat;

	       if (Sys()->FS()->IsFile (abs_mat_name))
	       {
		  cache->Get (V_MATERIAL, abs_mat_name, mat);
	       }
	       else
	       {
		   std::ostringstream os;
		   os << name << '?' << i;
		   String textureName( os.str() );

		   /// Create a new entry in the cache
		   cache->Get (V_MATERIAL, textureName, mat);
		   mat->m_Flags = MATERIAL_HAS_LIGHTING | MATERIAL_HAS_PASS1;
		   mat->m_Ambient = Color (1.0f, 1.0f, 1.0f);
		   mat->m_Diffuse = Color (1.0f, 1.0f, 1.0f);
		   mat->m_Specular = Color (0.0f, 0.0f, 0.0f);

		   TexturePtr texture;
		   if (cache->Get (V_TEXTURE, textureName, texture))
		   {
		       texture->Configure();

		       mat->m_Passes[0].m_Texture = texture;
		       mat->m_Passes[0].m_Flags |= PASS_HAS_TEXTURE;
		   }
	       }
	       
	       mdl->m_Skin.Insert(mat);
	    }

	    // Copy attachments ==========================
	    mstudioattachment_t *pattach;
	    pattach = (mstudioattachment_t *)(pin + hdr->attachmentindex);

	    /// Copy skins..
	    for (int j = 0; j < hdr->numattachments; j++)
	    {
	       if (pattach[j].name[0] == '\0')
	       {
		  sprintf (pattach[j].name, "%d", j);
	       }

	       Attachment &attach = mdl->m_Attachments[pattach[j].name];

	       attach.m_Bone = pattach[j].bone;
	       attach.m_Origin = pattach[j].org;
	       attach.m_Vectors[0] = pattach[j].vectors[0];
	       attach.m_Vectors[1] = pattach[j].vectors[1];
	       attach.m_Vectors[2] = pattach[j].vectors[2];

#if 0
	       Sys()->Log ("Attachment '%s':\n"
			   "\tbone %d\n"
			   "\torg %3.2f %3.2f %3.2f\n"
			   "\tv0 %3.2f %3.2f %3.2f\n"
			   "\tv1 %3.2f %3.2f %3.2f\n"
			   "\tv2 %3.2f %3.2f %3.2f\n",
			   pattach[j].name,
			   pattach[j].bone,
			   pattach[j].org.X,
			   pattach[j].org.Y,
			   pattach[j].org.Z,
			   pattach[j].vectors[0].X,
			   pattach[j].vectors[0].Y,
			   pattach[j].vectors[0].Z,
			   pattach[j].vectors[1].X,
			   pattach[j].vectors[1].Y,
			   pattach[j].vectors[1].Z,
			   pattach[j].vectors[2].X,
			   pattach[j].vectors[2].Y,
			   pattach[j].vectors[2].Z);
#endif
	    }


	    // Copy models.. ==============================
	    mdl->m_SubModels.resize (hdr->numbodyparts);
	    
	    for (int b = 0; b < hdr->numbodyparts; b++)
	    {
	       mstudiobodyparts_t *bodypart =
		  (mstudiobodyparts_t *)
		  ((uchar *)hdr + hdr->bodypartindex) + b;
	       mstudiomodel_t *pmodel =
		  (mstudiomodel_t *)
		  ((uchar *)hdr + bodypart->modelindex);

	       SubModel &sm = mdl->m_SubModels[b];
      
	       // Prepare vertices.. ==============================
	       sm.Builder()->SetFormat
		  (VertexBuffer::VB_HAS_COORD | VertexBuffer::VB_HAS_UV0 | VertexBuffer::VB_HAS_NORMAL);
	       
	       Vector3 *coords = (Vector3*) ((uchar*) hdr + pmodel->vertindex);
	       uchar *boneinfos = ((uchar*) hdr + pmodel->vertinfoindex);

	       Vector3 *normals = (Vector3*)
		  ((uchar*) hdr + pmodel->normindex);
	       //uchar *normboneinfos = ((uchar*) hdr + pmodel->norminfoindex);
	       

	       for (int i = 0; i < pmodel->nummesh; i++)
	       {
		  mstudiomesh_t *pmesh =
		     (mstudiomesh_t *)((uchar *)hdr + pmodel->meshindex) + i;
		  short *tricmds = (short *)((uchar *)hdr + pmesh->triindex);
		  short *ptricmds = tricmds;
		  
		  int skin = pmesh->skinref;

		  // Try to get texture size...
		  Texture *tex = &*mdl->m_Skin.GetMaterial(skin)->m_Passes[0].m_Texture;
		  float w, h;
		  if (tex)
		  {
		     w = static_cast<scalar>(tex->m_Width);
		     h = static_cast<scalar>(tex->m_Height);
		  }
		  else
		  {
		     w = static_cast<scalar>(ptexture[skin].width);
		     h = static_cast<scalar>(ptexture[skin].height);
		  }

		  /// Count triangle commands
		  int ntri = 0, n;
		  while ((n = (*ptricmds++)) != 0)
		  {
		     ntri++;
		     if (n < 0)
			n = -n;

		     ptricmds += 4 * n;
		     ntri += 4 * n;
		  }

		  /// Copy triangle commands
		  ptricmds = tricmds;
		  while ((n = (*ptricmds++)) != 0)
		  {
		     PrimitiveType type;

		     if (n < 0)
		     {
			n = -n;
			type = PRIM_TRIANGLE_FAN;
		     }
		     else
			type = PRIM_TRIANGLE_STRIP;
		     
		     /// Fill vertex array
		     std::vector<Vertex> vertices(n);
		     PrimitiveBlock pb;

		     pb.SetType (type);
		     pb.Resize (n);

		     for (int i = 0; i < n; i++)
		     {
			short nvert = ptricmds[i * 4];
			short nnorm = ptricmds[i * 4+1];
			short us = ptricmds[i * 4+2];
			short vs = ptricmds[i * 4+3];

			vertices[i].m_Coord = coords[nvert];
			vertices[i].m_Normal = normals[nnorm];
			vertices[i].m_TexCoord = Vector2
			   (float(us)/w, float(vs)/h);
			vertices[i].m_Bone = boneinfos[nvert];

			pb[i] = i;
		     }

		     sm.Builder()->SetVertices (&vertices[0], n);
		     sm.Builder()->AddBlock (skin, pb);

		     ptricmds += 4 * n;
		  }
	       }

	       sm.Build();
	    }

	    /// Some transformations have to be made in order to have
	    /// Y -> top.. (dirty hack, as most of this file :)
	    Matrix44 m; 
		m.MakeRotationX(-90.f);

	    mdl->SetupBBox (false);
	    mdl->m_BBox.Transform (m);
	    skel->m_RootMatrix = m;

	    delete hdr;

	    scalar height;	    
	    if ( StringToScalar(args, height) )
	    {
	       ((Object*)skel)->PostLoad(cache); // uh uh trick to get a valid
	       // skel->m_Cache pointer.
	       SequencePtr seq = skel->GetSequence("walk");
	       scalar scale = 
		  height / (seq->m_BBox.m_Max.Y - seq->m_BBox.m_Min.Y);

	       if (scale != 0.0)
		  mdl->Scale (scale);
	    }

	    return mdl != 0;
	 }  // Load()
	 
   };
   
   extern void ark_AddMdlLoader (Loaders *loaders)
   {
      loaders->Add (new LoaderMDL);
   }

} // namespace Ark


